# Add tests here

from typing import List
import unittest
import elara


class RunTests(unittest.TestCase):
    def setUp(self):
        self.db = elara.exe("test.db", False)
        self.db.hnew("newmap")

    def test_hnew(self):
        with self.assertRaises(Exception):
            self.db.hnew("key", -1)

    def test_hadd(self):
        self.assertEqual(len(self.db.get("newmap")), 0)
        self.db.hadd("newmap", "one", 1)
        self.assertEqual(len(self.db.get("newmap")), 1)

        self.assertEqual(self.db.hadd("newmap1", "two", 2), False)
        new_map = self.db.get("newmap")
        self.assertEqual(new_map["one"], 1)
        self.db.hadd("newmap", "string", "text")
        new_map = self.db.get("newmap")
        self.assertEqual(new_map["string"], "text")

    def test_haddt(self):
        self.db.haddt("newmap", ("one", 1))
        self.assertEqual(self.db.get("newmap")["one"], 1)
        self.db.haddt("newmap", ("two", 2))
        self.assertEqual(len(self.db.get("newmap")), 2)

    def test_hget(self):
        self.db.haddt("newmap", ("one", 1))
        self.assertEqual(self.db.hget("newmap", "one"), 1)
        self.assertEqual(self.db.hget("newmap1", "two"), False)

    def test_hpop(self):
        self.db.haddt("newmap", ("one", 1))
        self.db.haddt("newmap", ("two", 2))
        self.assertEqual(self.db.hpop("newmap", "one"), 1)
        self.assertEqual(self.db.hpop("newmap1", "one"), False)

    def test_hkeys(self):
        self.db.haddt("newmap", ("one", 1))
        self.db.haddt("newmap", ("two", 2))
        assert "one" in self.db.hkeys("newmap")
        assert len(self.db.hkeys("newmap")) == 2

    def test_hvals(self):
        self.db.haddt("newmap", ("one", "demo"))
        self.db.haddt("newmap", ("two", 2))
        assert "demo" in self.db.hvals("newmap")
        assert 2 in self.db.hvals("newmap")
        assert len(self.db.hvals("newmap")) == 2

    def test_hexists(self):
        self.db.haddt("newmap", ("one", "demo"))
        self.db.haddt("newmap", ("two", 2))
        self.assertEqual(self.db.hexists("newmap", "one"), True)
        self.assertEqual(self.db.hexists("newmap", "three"), False)
        self.assertEqual(self.db.hexists("newmap1", "one"), False)

    def test_hmerge(self):
        self.db.hadd("newmap", "one", 1)
        self.db.hadd("newmap", "two", 2)
        self.assertEqual(len(self.db.get("newmap")), 2)
        new_dict = {"two": 20, "three": 3}
        self.db.hmerge("newmap", new_dict)
        self.assertEqual(len(self.db.get("newmap")), 3)
        self.assertEqual(self.db.get("newmap")["two"], 20)



"""
Copyright (c) 2021, Saurabh Pujari
All rights reserved.

This source code is licensed under the BSD-style license found in the LICENSE file in the root directory of this source tree.
"""

import datetime
import math
from .status import Status


class Cache_obj:
    def __init__(self, key, max_age=None):
        self.key = key
        if max_age == None:
            self.expiry = None
        else:
            self.expiry = datetime.datetime.now() + datetime.timedelta(seconds=max_age)


class LRU:
    def __init__(self, max_size=math.inf) -> None:
        self.cache = []
        self.size = 0
        # max_size needs to be verified in Elara() to automatically call cull()
        self.max_size = max_size

    # Load LRU everytime DB is loaded
    # Set max_age to the default class max_age
    def _load(self, db, max_age):
        for key in db.keys():
            cache_obj = Cache_obj(key, max_age)
            self.push(cache_obj)

    def _get_cache_object(self, key):
        for i in range(0, self.size):
            if self.cache[i].key == key:
                return self.cache[i]
        return False

    def _resolve_size(self):
        if self.size < self.max_size:
            return True
        else:
            return Status.FULL

    def delete_if_expired(self, key):
        cache_obj = self._get_cache_object(key)

        if cache_obj == False:
            return Status.NOTFOUND
        if cache_obj.expiry == None:
            return False
        if cache_obj.expiry <= datetime.datetime.now():
            self.rem(cache_obj)
            return Status.EXPIRED
        else:
            return False

    # Push a new key into the cache
    # IF key already exists then overwrite it
    def push(self, new_cache_obj):
        cache_obj = self._get_cache_object(new_cache_obj.key)

        # If cache_obj is not present, resolve size before pushing
        if cache_obj == False:
            resolve = self._resolve_size()
            if resolve == Status.FULL:
                return Status.FULL
            else:
                self.cache.insert(0, new_cache_obj)
                self.size += 1
                return True
        # If its present, simply delete it and push the new_cache_obj
        else:
            self.rem(cache_obj)
            self.cache.insert(0, new_cache_obj)
            self.size += 1
            return True

    # Pop the least recently used key (end of the cache list)
    def pop(self):
        if self.size == 0:
            return False
        else:
            cache_obj = self.cache[-1]
            del self.cache[-1]
            self.size -= 1
            return cache_obj

    def rem_key(self, key):
        cache_obj = self._get_cache_object(key)
        if cache_obj == False:
            return Status.NOTFOUND
        self.rem(cache_obj)

    def rem(self, cache_obj):
        if cache_obj in self.cache:
            self.cache.remove(cache_obj)
            self.size -= 1
        else:
            return False

    # Peek the most recently used key (start of the cache list)
    def peek(self):
        if self.delete_if_expired(self.cache[0]) == Status.EXPIRED:
            return Status.EXPIRED
        else:
            return self.cache[0]

    # return a list of keys to delete and the remaining cache
    def all(self):
        deleted_keys = []
        for cache_obj in self.cache:
            key = cache_obj.key
            if self.delete_if_expired(cache_obj.key) == Status.EXPIRED:
                deleted_keys.append(key)
        return deleted_keys, self.cache

    # Bring key to the front of the cache
    # Take key as an argument and retrieve the cache_obj
    def touch(self, key):

        res = self.delete_if_expired(key)

        if res == Status.EXPIRED:
            return Status.EXPIRED
        elif res == Status.NOTFOUND:
            return Status.NOTFOUND

        cache_obj = self._get_cache_object(key)

        if self.cache[0] == cache_obj:
            return True
        else:
            self.cache.remove(cache_obj)
            self.size -= 1
            self.push(cache_obj)
            return True

    def clear(self):
        self.cache = []
        self.size = 0

    def print(self):
        print(self.cache)



"""
Copyright (c) 2021, Saurabh Pujari
All rights reserved.

This source code is licensed under the BSD-style license found in the LICENSE file in the root directory of this source tree.
"""
import os
from .elarautil import Util
from .lru import LRU, Cache_obj
from .status import Status
from .exceptions import InvalidCacheParams


def is_pos(val):
    return isinstance(val, int) and val > 0


class Elara:

    from .strings import setnx, append, getset, mget, mset, msetnx, slen
    from .lists import (
        lnew,
        lpush,
        lextend,
        lindex,
        lrange,
        lrem,
        lpop,
        llen,
        lappend,
        lexists,
        linsert,
    )
    from .hashtables import hnew, hadd, haddt, hget, hpop, hkeys, hvals, hexists, hmerge
    from .shared import (
        retmem,
        retdb,
        retkey,
        commit,
        exportdb,
        exportkeys,
        exportmem,
        securedb,
        updatekey,
    )

    def __init__(self, path, commitdb, key_path=None, cache_param=None):
        self.path = os.path.expanduser(path)
        self.commitdb = commitdb

        if cache_param == None:
            self.lru = LRU()
            self.max_age = None
            self.cull_freq = 20  # Delete 20% by default
        else:  # exe_cache() mode
            if "max_age" in cache_param and "max_size" in cache_param:
                if is_pos(cache_param["max_age"]) and is_pos(cache_param["max_size"]):
                    self.lru = LRU(cache_param["max_size"])
                    self.max_age = cache_param["max_age"]
                else:
                    raise InvalidCacheParams("max_age, max_size")
            elif "max_age" in cache_param:
                if is_pos(cache_param["max_age"]):
                    self.max_age = cache_param["max_age"]
                    self.lru = LRU()
                else:
                    raise InvalidCacheParams("max_age")
            elif "max_size" in cache_param:
                if is_pos(cache_param["max_size"]):
                    self.lru = LRU(cache_param["max_size"])
                    self.max_age = None
                else:
                    raise InvalidCacheParams("max_size")
            if "cull_freq" in cache_param:
                if is_pos(cache_param["cull_freq"]) and cache_param["cull_freq"] <= 100:
                    self.cull_freq = cache_param["cull_freq"]
                else:
                    raise InvalidCacheParams("cull_freq")
            else:
                self.cull_freq = 20

        # this is in place to prevent opening incompatible databases between versions of the storage format
        self.db_format_version = 0x0001

        # Since key file is generated first, invalid token error for pre existing open dbs
        # Load the database key
        if not key_path == None:
            new_key_path = os.path.expanduser(key_path)
            if os.path.exists(new_key_path):
                file = open(new_key_path, "rb")
                self.key = file.read()
                file.close()
            else:
                self.key = None
        else:
            self.key = None

        # Load the data
        if os.path.exists(path):
            self._load()
        else:
            self.db = {}

    def _load(self):
        if self.key:
            self.db = Util.read_and_decrypt(self)
            self.lru._load(self.db, self.max_age)
        else:
            self.db = Util.read_plain_db(self)
            self.lru._load(self.db, self.max_age)

    def _dump(self):
        if self.key:
            Util.encrypt_and_store(self)  # Enclose in try-catch
        else:
            Util.store_plain_db(self)

    def _autocommit(self):
        if self.commitdb:
            self._dump()

    # remove a list of keys from db only
    def _remkeys_db_only(self, keys=[]):
        for key in keys:
            del self.db[key]
        self._autocommit()

    # Take max_age or self.max_age
    def set(self, key, value, max_age=None):
        if isinstance(key, str):
            if max_age == None:
                cache_obj = Cache_obj(key, self.max_age)
            elif max_age == "i":
                cache_obj = Cache_obj(key, None)
            else:
                if is_pos(max_age):
                    cache_obj = Cache_obj(key, max_age)
                else:
                    raise InvalidCacheParams("max_age")

            # this is for when a key is being overwritten by set
            if self.exists(key):
                self.rem(key)
                self.lru.push(cache_obj)
            elif self.lru.push(cache_obj) == Status.FULL:
                self.cull(self.cull_freq)
                self.lru.push(cache_obj)

            self.db[key] = value
            self._autocommit()
            return True
        else:
            raise Exception

    def get(self, key):
        try:
            res = self.lru.touch(key)
            if res == Status.EXPIRED:
                del self.db[key]
                return None
            elif res == Status.NOTFOUND:
                return None
            else:
                return self.db[key]
        except KeyError:
            return None

    def rem(self, key):
        if self.lru.rem_key(key) == Status.NOTFOUND:
            raise Exception
        del self.db[key]
        self._autocommit()
        return True

    def remkeys(self, keys=[]):
        for key in keys:
            if self.exists(key):
                self.rem(key)
            self._autocommit()

    def clear(self):
        self.lru.clear()
        self.db = {}
        self._autocommit()
        return True

    def exists(self, key):
        res = self.lru.touch(key)
        if res == Status.EXPIRED:
            del self.db[key]
            self._autocommit()
            return False
        elif res == Status.NOTFOUND:
            return False
        return key in self.db

    def cull(self, percentage=20):
        if 0 <= percentage <= 100:
            count = int((percentage / 100) * (self.lru.size))
            # print("final count", count)

            if count == 0 and (percentage > 0 and self.lru.size > 0):
                cache_obj = self.lru.pop()
                del self.db[cache_obj.key]
            else:
                for i in range(0, count):
                    cache_obj = self.lru.pop()
                    if cache_obj == False:
                        break
                    del self.db[cache_obj.key]

            self._autocommit()
            return True
        else:
            return False

    def getkeys(self):
        deleted_keys, cache = self.lru.all()
        self._remkeys_db_only(deleted_keys)

        keys = []
        for cache_obj in cache:
            keys.append(cache_obj.key)

        return keys

    def numkeys(self):
        deleted_keys, cache = self.lru.all()
        self._remkeys_db_only(deleted_keys)

        return len(cache)

    def incr(self, key, val=1):
        if self.exists(key):
            data = self.get(key)
            if isinstance(data, (int, float)):
                data += val
                data = round(data, 3)
                self.set(key, data)
            else:
                return False
        else:
            return False

    def decr(self, key, val=1):
        if self.exists(key):
            data = self.get(key)
            if isinstance(data, (int, float)):
                data -= val
                data = round(data, 3)
                self.set(key, data)
            else:
                return False
        else:
            return False



# Add some tests here

import os
import unittest

import elara


class RunTests(unittest.TestCase):
    def setUp(self):
        self.db = elara.exe("test.db", False)

    def test_exe(self):
        res = elara.exe("test.db", False)
        assert res is not None

    def test_store_restore_data(self):
        db = elara.exe("test_store.db")
        db.set(
            "test_key",
            'test_data:"ði ıntəˈnæʃənəl fəˈnɛtık əsoʊsiˈeıʃn Y [ˈʏpsilɔn], Yen [jɛn], Yoga [ˈjoːgɑ]"',
        )
        db.commit()
        db_load = elara.exe("test_store.db")
        recov_data = db_load.get("test_key")
        self.assertEqual(
            recov_data,
            'test_data:"ði ıntəˈnæʃənəl fəˈnɛtık əsoʊsiˈeıʃn Y [ˈʏpsilɔn], Yen [jɛn], Yoga [ˈjoːgɑ]"',
        )
        # cleanup database files
        os.remove("test_store.db")

    def test_store_restore_data_secure(self):
        db = elara.exe_secure("test_enc.db")
        db.set(
            "test_key",
            'test_data:"ði ıntəˈnæʃənəl fəˈnɛtık əsoʊsiˈeıʃn Y [ˈʏpsilɔn], Yen [jɛn], Yoga [ˈjoːgɑ]"',
        )
        db.commit()
        db_load = elara.exe_secure("test_enc.db")
        recov_data = db_load.get("test_key")
        self.assertEqual(
            recov_data,
            'test_data:"ði ıntəˈnæʃənəl fəˈnɛtık əsoʊsiˈeıʃn Y [ˈʏpsilɔn], Yen [jɛn], Yoga [ˈjoːgɑ]"',
        )

        # cleanup database files
        os.remove("test_enc.db")
        os.remove("edb.key")

    def test_atomicity(self):
        class Func:
            def __init__(self, name):
                self.name = name

        db = elara.exe("test_atomicity.db", False)
        obj = Func("test")

        with self.assertRaises(Exception):
            db.set("key", obj)
            db.commit()



"""
Copyright (c) 2021, Saurabh Pujari
All rights reserved.

This source code is licensed under the BSD-style license found in the LICENSE file in the root directory of this source tree.
"""
from .elara import Elara
from .elarautil import Util
from .lru import LRU, Cache_obj


def exe(path, commitdb=False):
    return Elara(path, commitdb, None, None)


def exe_secure(path, commitdb=False, key_path="edb.key"):
    Util.loadkey(key_path)
    return Elara(path, commitdb, key_path, None)


def exe_cache(path, cache_param=None, commitdb=False):
    return Elara(path, commitdb, None, cache_param)
